@@ -16,6 +16,7 @@ class FakeArrayImplementation(BaseConcreteArray):
         self.dtype = base.get_dtype()
         self.shape = [base.get_size()]
         self.storage = self._base.implementation.storage
+        self.order = base.get_order()
 
     def base(self):
         return self._base
@@ -97,7 +98,7 @@ class W_FlatIterator(W_NDimArray):
         finally:
             self.iter.reset(self.state, mutate=True)
 
-    def descr___array_wrap__(self, space, obj):
+    def descr___array_wrap__(self, space, obj, w_context=None):
         return obj
 
 W_FlatIterator.typedef = TypeDef("numpy.flatiter",
