@@ -1,26 +1,28 @@
 from rpython.jit.metainterp.walkvirtual import VirtualVisitor
-from rpython.jit.metainterp.history import (BoxInt, ConstInt, BoxPtr, Const,
-        ConstPtr, ConstFloat)
-from rpython.jit.metainterp.optimizeopt import virtualize
-from rpython.jit.metainterp.optimizeopt.intutils import IntUnbounded
-from rpython.jit.metainterp.optimizeopt.optimizer import (LEVEL_CONSTANT,
-    LEVEL_KNOWNCLASS, LEVEL_NONNULL, LEVEL_UNKNOWN, OptValue)
-from rpython.jit.metainterp.resoperation import rop, ResOperation
-from rpython.rlib.debug import debug_start, debug_stop, debug_print
-from rpython.rlib.objectmodel import we_are_translated, import_from_mixin
+from rpython.jit.metainterp.history import ConstInt, ConstPtr, ConstFloat
+from rpython.jit.metainterp.optimizeopt.info import ArrayPtrInfo,\
+     ArrayStructInfo, AbstractStructPtrInfo
+from rpython.jit.metainterp.optimizeopt.intutils import \
+     MININT, MAXINT, IntBound, IntLowerBound
+from rpython.jit.metainterp.resoperation import rop, ResOperation,\
+     InputArgInt, InputArgRef, InputArgFloat
+from rpython.rlib.debug import debug_print
 
-
-class BadVirtualState(Exception):
-    pass
+LEVEL_UNKNOWN = '\x00'
+LEVEL_NONNULL = '\x01'
+LEVEL_KNOWNCLASS = '\x02'
+LEVEL_CONSTANT = '\x03'
 
 class VirtualStatesCantMatch(Exception):
     def __init__(self, msg='?', state=None):
         self.msg = msg
         self.state = state
 
+
 class GenerateGuardState(object):
-    def __init__(self, cpu=None, guards=None, renum=None, bad=None):
-        self.cpu = cpu
+    def __init__(self, optimizer=None, guards=None, renum=None, bad=None):
+        self.optimizer = optimizer
+        self.cpu = optimizer.cpu
         if guards is None:
             guards = []
         self.extra_guards = guards
@@ -31,20 +33,50 @@ class GenerateGuardState(object):
             bad = {}
         self.bad = bad
 
+    def get_runtime_item(self, box, descr, i):
+        array = box.getref_base()
+        if descr.is_array_of_pointers():
+            return InputArgRef(self.cpu.bh_getarrayitem_gc_r(array, i, descr))
+        elif descr.is_array_of_floats():
+            return InputArgFloat(self.cpu.bh_getarrayitem_gc_f(array, i, descr))
+        else:
+            return InputArgInt(self.cpu.bh_getarrayitem_gc_i(array, i, descr))
+
+    def get_runtime_field(self, box, descr):
+        struct = box.getref_base()
+        if descr.is_pointer_field():
+            return InputArgRef(self.cpu.bh_getfield_gc_r(struct, descr))
+        elif descr.is_float_field():
+            return InputArgFloat(self.cpu.bh_getfield_gc_f(struct, descr))
+        else:
+            return InputArgInt(self.cpu.bh_getfield_gc_i(struct, descr))
+
+    def get_runtime_interiorfield(self, box, descr, i):
+        struct = box.getref_base()
+        if descr.is_pointer_field():
+            return InputArgRef(self.cpu.bh_getinteriorfield_gc_r(struct, i,
+                                                                 descr))
+        elif descr.is_float_field():
+            return InputArgFloat(self.cpu.bh_getinteriorfield_gc_f(struct, i,
+                                                                   descr))
+        else:
+            return InputArgInt(self.cpu.bh_getinteriorfield_gc_i(struct, i,
+                                                                 descr))
+
 class AbstractVirtualStateInfo(object):
     position = -1
 
-    def generate_guards(self, other, value, state):
+    def generate_guards(self, other, op, runtime_op, state):
         """ generate guards (output in the list extra_guards) that make runtime
         values of the shape other match the shape of self. if that's not
         possible, VirtualStatesCantMatch is thrown and bad gets keys set which
         parts of the state are the problem.
 
-        the function can peek into value (and particularly also the boxes in
-        the value) as a guiding heuristic whether making such guards makes
-        sense. if None is passed in for value, no guard is ever generated, and
+        the function can peek into the information about the op, as well
+        as runtime value (passed in runtime_op)
+        as a guiding heuristic whether making such guards makes
+        sense. if None is passed in for op, no guard is ever generated, and
         this function degenerates to a generalization check."""
-        assert value is None or isinstance(value, OptValue)
         assert self.position != -1
         if self.position in state.renum:
             if state.renum[self.position] != other.position:
@@ -58,19 +90,19 @@ class AbstractVirtualStateInfo(object):
         else:
             state.renum[self.position] = other.position
             try:
-                self._generate_guards(other, value, state)
+                self._generate_guards(other, op, runtime_op, state)
             except VirtualStatesCantMatch, e:
                 state.bad[self] = state.bad[other] = None
                 if e.state is None:
                     e.state = state
                 raise e
 
-    def _generate_guards(self, other, value, state):
+    def _generate_guards(self, other, box, runtime_box, state):
         raise VirtualStatesCantMatch(
                 'Generating guards for making the VirtualStates ' +
                 'at hand match have not been implemented')
 
-    def enum_forced_boxes(self, boxes, value, optimizer):
+    def enum_forced_boxes(self, boxes, box, optimizer, force_boxes=False):
         raise NotImplementedError
 
     def enum(self, virtual_state):
@@ -103,16 +135,19 @@ class AbstractVirtualStructStateInfo(AbstractVirtualStateInfo):
     def __init__(self, fielddescrs):
         self.fielddescrs = fielddescrs
 
-    def _generate_guards(self, other, value, state):
+    def _generate_guards(self, other, box, runtime_box, state):
         if not self._generalization_of_structpart(other):
             raise VirtualStatesCantMatch("different kinds of structs")
 
         assert isinstance(other, AbstractVirtualStructStateInfo)
         assert len(self.fielddescrs) == len(self.fieldstate)
         assert len(other.fielddescrs) == len(other.fieldstate)
-        if value is not None:
-            assert isinstance(value, virtualize.AbstractVirtualStructValue)
-            assert value.is_virtual()
+        if runtime_box is not None:
+            opinfo = state.optimizer.getptrinfo(box)
+            assert opinfo.is_virtual()
+            assert isinstance(opinfo, AbstractStructPtrInfo)
+        else:
+            opinfo = None
 
         if len(self.fielddescrs) != len(other.fielddescrs):
             raise VirtualStatesCantMatch("field descrs don't match")
@@ -120,36 +155,50 @@ class AbstractVirtualStructStateInfo(AbstractVirtualStateInfo):
         for i in range(len(self.fielddescrs)):
             if other.fielddescrs[i] is not self.fielddescrs[i]:
                 raise VirtualStatesCantMatch("field descrs don't match")
-            if value is not None:
-                v = value._fields[self.fielddescrs[i]] # must be there
+            if runtime_box is not None and opinfo is not None:
+                fieldbox = opinfo._fields[self.fielddescrs[i].get_index()]
+                # must be there
+                fieldbox_runtime = state.get_runtime_field(runtime_box,
+                                                           self.fielddescrs[i])
             else:
-                v = None
-            self.fieldstate[i].generate_guards(other.fieldstate[i], v, state)
+                fieldbox = None
+                fieldbox_runtime = None
+            if self.fieldstate[i] is not None:
+                if other.fieldstate[i] is None:
+                    raise VirtualStatesCantMatch
+                self.fieldstate[i].generate_guards(other.fieldstate[i],
+                                                   fieldbox,
+                                                   fieldbox_runtime, state)
 
 
     def _generalization_of_structpart(self, other):
         raise NotImplementedError
 
-    def enum_forced_boxes(self, boxes, value, optimizer):
-        if not isinstance(value, virtualize.AbstractVirtualStructValue):
-            raise BadVirtualState
-        if not value.is_virtual():
-            raise BadVirtualState
+    def enum_forced_boxes(self, boxes, box, optimizer, force_boxes=False):
+        box = optimizer.get_box_replacement(box)
+        info = optimizer.getptrinfo(box)
+        if info is None or not info.is_virtual():
+            raise VirtualStatesCantMatch()
+        else:
+            assert isinstance(info, AbstractStructPtrInfo)
         for i in range(len(self.fielddescrs)):
-            try:
-                v = value._fields[self.fielddescrs[i]]
-            except KeyError:
-                raise BadVirtualState
-            s = self.fieldstate[i]
-            if s.position > self.position:
-                s.enum_forced_boxes(boxes, v, optimizer)
+            state = self.fieldstate[i]
+            if not state:
+                continue
+            if state.position > self.position:
+                fieldbox = info._fields[i]
+                state.enum_forced_boxes(boxes, fieldbox, optimizer, force_boxes)
 
     def _enum(self, virtual_state):
         for s in self.fieldstate:
-            s.enum(virtual_state)
+            if s:
+                s.enum(virtual_state)
 
 
 class VirtualStateInfo(AbstractVirtualStructStateInfo):
+    def is_virtual(self):
+        return True
+
     def __init__(self, known_class, fielddescrs):
         AbstractVirtualStructStateInfo.__init__(self, fielddescrs)
         self.known_class = known_class
@@ -181,50 +230,58 @@ class VArrayStateInfo(AbstractVirtualStateInfo):
     def __init__(self, arraydescr):
         self.arraydescr = arraydescr
 
-    def _generate_guards(self, other, value, state):
+    def _generate_guards(self, other, box, runtime_box, state):
         if not isinstance(other, VArrayStateInfo):
             raise VirtualStatesCantMatch("other is not an array")
         if self.arraydescr is not other.arraydescr:
             raise VirtualStatesCantMatch("other is a different kind of array")
         if len(self.fieldstate) != len(other.fieldstate):
             raise VirtualStatesCantMatch("other has a different length")
-        v = None
+        fieldbox = None
+        fieldbox_runtime = None
         for i in range(len(self.fieldstate)):
-            if value is not None:
-                assert isinstance(value, virtualize.VArrayValue)
-                v = value._items[i]
-            self.fieldstate[i].generate_guards(other.fieldstate[i],
-                                               v, state)
+            if runtime_box is not None:
+                opinfo = state.optimizer.getptrinfo(box)
+                assert isinstance(opinfo, ArrayPtrInfo)
+                fieldbox = opinfo._items[i]
+                fieldbox_runtime = state.get_runtime_item(runtime_box,
+                                            self.arraydescr, i)
+            if self.fieldstate[i] is not None:
+                if other.fieldstate[i] is None:
+                    raise VirtualStatesCantMatch
+                self.fieldstate[i].generate_guards(other.fieldstate[i],
+                                            fieldbox, fieldbox_runtime, state)
 
-    def enum_forced_boxes(self, boxes, value, optimizer):
-        if not isinstance(value, virtualize.VArrayValue):
-            raise BadVirtualState
-        if not value.is_virtual():
-            raise BadVirtualState
-        if len(self.fieldstate) > value.getlength():
-            raise BadVirtualState
+    def enum_forced_boxes(self, boxes, box, optimizer, force_boxes=False):
+        box = optimizer.get_box_replacement(box)
+        info = optimizer.getptrinfo(box)
+        if info is None or not info.is_virtual():
+            raise VirtualStatesCantMatch()
+        if len(self.fieldstate) > info.getlength():
+            raise VirtualStatesCantMatch
         for i in range(len(self.fieldstate)):
-            v = value.get_item_value(i)
-            if v is None:
-                v = value.get_missing_null_value()
+            fieldbox = info.getitem(self.arraydescr, i)
             s = self.fieldstate[i]
-            if s.position > self.position:
-                s.enum_forced_boxes(boxes, v, optimizer)
+            if s is not None:
+                if s.position > self.position:
+                    s.enum_forced_boxes(boxes, fieldbox, optimizer, force_boxes)
 
     def _enum(self, virtual_state):
         for s in self.fieldstate:
-            s.enum(virtual_state)
+            if s:
+                s.enum(virtual_state)
 
     def debug_header(self, indent):
         debug_print(indent + 'VArrayStateInfo(%d):' % self.position)
 
 
 class VArrayStructStateInfo(AbstractVirtualStateInfo):
-    def __init__(self, arraydescr, fielddescrs):
+    def __init__(self, arraydescr, fielddescrs, length):
         self.arraydescr = arraydescr
         self.fielddescrs = fielddescrs
+        self.length = length
 
-    def _generate_guards(self, other, value, state):
+    def _generate_guards(self, other, box, runtime_box, state):
         if not isinstance(other, VArrayStructStateInfo):
             raise VirtualStatesCantMatch("other is not an VArrayStructStateInfo")
         if self.arraydescr is not other.arraydescr:
@@ -233,72 +290,106 @@ class VArrayStructStateInfo(AbstractVirtualStateInfo):
         if len(self.fielddescrs) != len(other.fielddescrs):
             raise VirtualStatesCantMatch("other has a different length")
 
-        p = 0
-        v = None
-        for i in range(len(self.fielddescrs)):
-            if len(self.fielddescrs[i]) != len(other.fielddescrs[i]):
-                raise VirtualStatesCantMatch("other has a different length")
-            for j in range(len(self.fielddescrs[i])):
-                descr = self.fielddescrs[i][j]
-                if descr is not other.fielddescrs[i][j]:
-                    raise VirtualStatesCantMatch("other is a different kind of array")
-                if value is not None:
-                    assert isinstance(value, virtualize.VArrayStructValue)
-                    v = value._items[i][descr]
-                self.fieldstate[p].generate_guards(other.fieldstate[p],
-                                                   v,
-                                                   state)
-                p += 1
+        if len(self.fielddescrs) != len(other.fielddescrs):
+            raise VirtualStatesCantMatch("other has a different length")
+        for j, descr in enumerate(self.fielddescrs):
+            if descr is not other.fielddescrs[j]:
+                raise VirtualStatesCantMatch("other is a different kind of array")
+        fieldbox = None
+        fieldbox_runtime = None
+        if box is not None:
+            opinfo = state.optimizer.getptrinfo(box)
+            assert isinstance(opinfo, ArrayPtrInfo)
+        else:
+            opinfo = None            
+        for i in range(self.length):
+            for descr in self.fielddescrs:
+                index = i * len(self.fielddescrs) + descr.get_index()
+                fieldstate = self.fieldstate[index]
+                if fieldstate is None:
+                    continue
+                if other.fieldstate[index] is None:
+                    raise VirtualStatesCantMatch
+                if box is not None and opinfo is not None:
+                    fieldbox = opinfo._items[index]
+                    fieldbox_runtime = state.get_runtime_interiorfield(
+                        runtime_box, descr, i)
+                self.fieldstate[index].generate_guards(other.fieldstate[index],
+                                       fieldbox, fieldbox_runtime, state)
 
     def _enum(self, virtual_state):
         for s in self.fieldstate:
-            s.enum(virtual_state)
+            if s is not None:
+                s.enum(virtual_state)
 
-    def enum_forced_boxes(self, boxes, value, optimizer):
-        if not isinstance(value, virtualize.VArrayStructValue):
-            raise BadVirtualState
-        if not value.is_virtual():
-            raise BadVirtualState
-        if len(self.fielddescrs) > len(value._items):
-            raise BadVirtualState
-        p = 0
-        for i in range(len(self.fielddescrs)):
-            for j in range(len(self.fielddescrs[i])):
-                try:
-                    v = value._items[i][self.fielddescrs[i][j]]
-                except KeyError:
-                    raise BadVirtualState
-                s = self.fieldstate[p]
-                if s.position > self.position:
-                    s.enum_forced_boxes(boxes, v, optimizer)
-                p += 1
+    def enum_forced_boxes(self, boxes, box, optimizer, force_boxes=False):
+        opinfo = optimizer.getptrinfo(box)
+        if not isinstance(opinfo, ArrayStructInfo):
+            raise VirtualStatesCantMatch
+        if not opinfo.is_virtual():
+            raise VirtualStatesCantMatch
+        #if len(self.fielddescrs) > len(value._items):
+        #    raise VirtualStatesCantMatch
+        for i in range(self.length):
+            for descr in self.fielddescrs:
+                index = i * len(self.fielddescrs) + descr.get_index()
+                fieldstate = self.fieldstate[index]
+                itembox = opinfo._items[i * len(self.fielddescrs) +
+                                        descr.get_index()]
+                if fieldstate is None:
+                    if itembox is not None:
+                        raise VirtualStatesCantMatch
+                    continue
+                # I think itembox must be present here
+                if fieldstate.position > self.position:
+                    fieldstate.enum_forced_boxes(boxes, itembox, optimizer,
+                                                 force_boxes)
 
     def debug_header(self, indent):
         debug_print(indent + 'VArrayStructStateInfo(%d):' % self.position)
 
-
 class NotVirtualStateInfo(AbstractVirtualStateInfo):
-    def __init__(self, value, is_opaque=False):
-        self.is_opaque = is_opaque
-        self.known_class = value.get_known_class()
-        self.level = value.getlevel()
-        if value.getintbound() is None:
-            self.intbound = IntUnbounded()
-        else:
-            self.intbound = value.getintbound().clone()
-        if value.is_constant():
-            self.constbox = value.box
-        else:
-            self.constbox = None
-        self.position_in_notvirtuals = -1
-        self.lenbound = value.getlenbound()
+    lenbound = None
+    intbound = None
+    level = LEVEL_UNKNOWN
+    constbox = None
+    known_class = None
+    
+    def __init__(self, cpu, type, info):
+        if info and info.is_constant():
+            self.level = LEVEL_CONSTANT
+            self.constbox = info.getconst()
+            if type == 'r':
+                self.known_class = info.get_known_class(cpu)
+            elif type == 'i':
+                self.intbound = info
+        elif type == 'r':
+            if info:
+                self.known_class = info.get_known_class(cpu)
+                if self.known_class:
+                    self.level = LEVEL_KNOWNCLASS
+                elif info.is_nonnull():
+                    self.level = LEVEL_NONNULL
+                self.lenbound = info.getlenbound(None)
+        elif type == 'i':
+            if isinstance(info, IntBound):
+                if info.lower < MININT / 2:
+                    info.lower = MININT
+                if info.upper > MAXINT / 2:
+                    info.upper = MAXINT
+                self.intbound = info
+        elif type == 'f':
+            if info and info.is_constant():
+                self.level = LEVEL_CONSTANT
+                self.constbox = info.getconst()
 
+    def is_const(self):
+        return self.constbox is not None
 
-    def _generate_guards(self, other, value, state):
-        if value is None or self.is_opaque:
-            box = None # generating guards for opaque pointers isn't safe
-        else:
-            box = value.box
+    def is_virtual(self):
+        return False
+
+    def _generate_guards(self, other, box, runtime_box, state):
         # XXX This will always retrace instead of forcing anything which
         # might be what we want sometimes?
         if not isinstance(other, NotVirtualStateInfo):
@@ -310,13 +401,20 @@ class NotVirtualStateInfo(AbstractVirtualStateInfo):
 
         extra_guards = state.extra_guards
         cpu = state.cpu
-        if self.lenbound and not self.lenbound.generalization_of(other.lenbound):
-            raise VirtualStatesCantMatch("length bound does not match")
+        if self.lenbound:
+            if other.lenbound is None:
+                other_bound = IntLowerBound(0)
+            else:
+                other_bound = other.lenbound
+            if not self.lenbound.contains_bound(other_bound):
+                raise VirtualStatesCantMatch("length bound does not match")
 
         if self.level == LEVEL_UNKNOWN:
             # confusingly enough, this is done also for pointers
             # which have the full range as the "bound", so it always works
-            return self._generate_guards_intbounds(other, box, extra_guards)
+            return self._generate_guards_intbounds(other, box, runtime_box,
+                                                   extra_guards,
+                                                   state.optimizer)
 
         # the following conditions often peek into the runtime value that the
         # box had when tracing. This value is only used as an educated guess.
@@ -325,7 +423,7 @@ class NotVirtualStateInfo(AbstractVirtualStateInfo):
         # will always generate correct behaviour, but performance will differ.
         elif self.level == LEVEL_NONNULL:
             if other.level == LEVEL_UNKNOWN:
-                if box is not None and box.nonnull():
+                if runtime_box is not None and runtime_box.nonnull():
                     op = ResOperation(rop.GUARD_NONNULL, [box], None)
                     extra_guards.append(op)
                     return
@@ -344,15 +442,16 @@ class NotVirtualStateInfo(AbstractVirtualStateInfo):
 
         elif self.level == LEVEL_KNOWNCLASS:
             if other.level == LEVEL_UNKNOWN:
-                if (box and box.nonnull() and
-                        self.known_class.same_constant(cpu.ts.cls_of_box(box))):
+                if (runtime_box and runtime_box.nonnull() and
+              self.known_class.same_constant(cpu.ts.cls_of_box(runtime_box))):
                     op = ResOperation(rop.GUARD_NONNULL_CLASS, [box, self.known_class], None)
                     extra_guards.append(op)
                     return
                 else:
                     raise VirtualStatesCantMatch("other's class is unknown")
             elif other.level == LEVEL_NONNULL:
-                if box and self.known_class.same_constant(cpu.ts.cls_of_box(box)):
+                if (runtime_box and self.known_class.same_constant(
+                        cpu.ts.cls_of_box(runtime_box))):
                     op = ResOperation(rop.GUARD_CLASS, [box, self.known_class], None)
                     extra_guards.append(op)
                     return
@@ -376,7 +475,7 @@ class NotVirtualStateInfo(AbstractVirtualStateInfo):
                 if self.constbox.same_constant(other.constbox):
                     return
                 raise VirtualStatesCantMatch("different constants")
-            if box is not None and self.constbox.same_constant(box.constbox()):
+            if runtime_box is not None and self.constbox.same_constant(runtime_box.constbox()):
                 op = ResOperation(rop.GUARD_VALUE, [box, self.constbox], None)
                 extra_guards.append(op)
                 return
@@ -384,27 +483,33 @@ class NotVirtualStateInfo(AbstractVirtualStateInfo):
                 raise VirtualStatesCantMatch("other not constant")
         assert 0, "unreachable"
 
-    def _generate_guards_intbounds(self, other, box, extra_guards):
+    def _generate_guards_intbounds(self, other, box, runtime_box, extra_guards,
+                                   optimizer):
+        if self.intbound is None:
+            return
         if self.intbound.contains_bound(other.intbound):
             return
-        if (box is not None and isinstance(box, BoxInt) and
-                self.intbound.contains(box.getint())):
+        if (runtime_box is not None and
+            self.intbound.contains(runtime_box.getint())):
             # this may generate a few more guards than needed, but they are
             # optimized away when emitting them
-            self.intbound.make_guards(box, extra_guards)
+            self.intbound.make_guards(box, extra_guards, optimizer)
             return
         raise VirtualStatesCantMatch("intbounds don't match")
 
-    def enum_forced_boxes(self, boxes, value, optimizer):
+    def enum_forced_boxes(self, boxes, box, optimizer, force_boxes=False):
         if self.level == LEVEL_CONSTANT:
             return
         assert 0 <= self.position_in_notvirtuals
-        if optimizer:
-            box = value.force_box(optimizer)
-        else:
-            if value.is_virtual():
-                raise BadVirtualState
-            box = value.get_key_box()
+        if optimizer is not None:
+            box = optimizer.get_box_replacement(box)
+            if box.type == 'r':
+                info = optimizer.getptrinfo(box)
+                if info and info.is_virtual():
+                    if force_boxes:
+                        info.force_box(box, optimizer)
+                    else:
+                        raise VirtualStatesCantMatch
         boxes[self.position_in_notvirtuals] = box
 
     def _enum(self, virtual_state):
@@ -456,50 +561,57 @@ class VirtualState(object):
         self.info_counter = -1
         self.numnotvirtuals = 0
         for s in state:
-            s.enum(self)
+            if s:
+                s.enum(self)
 
-    def generalization_of(self, other, bad=None, cpu=None):
-        state = GenerateGuardState(cpu=cpu, bad=bad)
+    def generalization_of(self, other, optimizer):
+        state = GenerateGuardState(optimizer)
         assert len(self.state) == len(other.state)
         try:
             for i in range(len(self.state)):
-                self.state[i].generate_guards(other.state[i], None, state)
+                self.state[i].generate_guards(other.state[i], None, None, state)
         except VirtualStatesCantMatch:
             return False
         return True
 
-    def generate_guards(self, other, values, cpu):
-        assert len(self.state) == len(other.state) == len(values)
-        state = GenerateGuardState(cpu)
+    def generate_guards(self, other, boxes, runtime_boxes, optimizer):
+        assert (len(self.state) == len(other.state) == len(boxes) ==
+                len(runtime_boxes))
+        state = GenerateGuardState(optimizer)
         for i in range(len(self.state)):
-            self.state[i].generate_guards(other.state[i], values[i],
-                                          state)
+            self.state[i].generate_guards(other.state[i], boxes[i],
+                                          runtime_boxes[i], state)
         return state
 
-    def make_inputargs(self, values, optimizer, keyboxes=False):
+    def make_inputargs(self, inputargs, optimizer, force_boxes=False):
         if optimizer.optearlyforce:
             optimizer = optimizer.optearlyforce
-        assert len(values) == len(self.state)
-        inputargs = [None] * self.numnotvirtuals
+        assert len(inputargs) == len(self.state)
+        boxes = [None] * self.numnotvirtuals
 
         # We try twice. The first time around we allow boxes to be forced
         # which might change the virtual state if the box appear in more
         # than one place among the inputargs.
-        for i in range(len(values)):
-            self.state[i].enum_forced_boxes(inputargs, values[i], optimizer)
-        for i in range(len(values)):
-            self.state[i].enum_forced_boxes(inputargs, values[i], None)
+        if force_boxes:
+            for i in range(len(inputargs)):
+                self.state[i].enum_forced_boxes(boxes, inputargs[i], optimizer,
+                                                True)
+        for i in range(len(inputargs)):
+            self.state[i].enum_forced_boxes(boxes, inputargs[i], optimizer)
 
-        if keyboxes:
-            for i in range(len(values)):
-                if not isinstance(self.state[i], NotVirtualStateInfo):
-                    box = values[i].get_key_box()
-                    assert not isinstance(box, Const)
-                    inputargs.append(box)
+        return boxes
 
-        assert None not in inputargs
-
-        return inputargs
+    def make_inputargs_and_virtuals(self, inputargs, optimizer):
+        inpargs = self.make_inputargs(inputargs, optimizer)
+        # we append the virtuals here in case some stuff is proven
+        # to be not a virtual and there are getfields in the short preamble
+        # that will read items out of there
+        virtuals = []
+        for i in range(len(inputargs)):
+            if not isinstance(self.state[i], NotVirtualStateInfo):
+                virtuals.append(inputargs[i])
+            
+        return inpargs, virtuals
 
     def debug_print(self, hdr='', bad=None, metainterp_sd=None):
         if bad is None:
@@ -523,46 +635,52 @@ class VirtualStateConstructor(VirtualVisitor):
     def already_seen_virtual(self, keybox):
         return keybox in self.fieldboxes
 
-    def getvalue(self, box):
-        return self.optimizer.getvalue(box)
-
-    def state(self, box):
-        value = self.getvalue(box)
-        box = value.get_key_box()
-        try:
-            info = self.info[box]
-        except KeyError:
-            self.info[box] = info = value.visitor_dispatch_virtual_type(self)
-            if value.is_virtual():
-                flds = self.fieldboxes[box]
-                info.fieldstate = [self.state_or_none(b, value) for b in flds]
-        return info
-
-    def state_or_none(self, box, value):
+    def create_state_or_none(self, box, opt):
         if box is None:
-            box = value.get_missing_null_value().box
-        return self.state(box)
+            return None
+        return self.create_state(box, opt)
+
+    def create_state(self, box, opt):
+        box = opt.get_box_replacement(box)
+        try:
+            return self.info[box]
+        except KeyError:
+            pass
+        if box.type == 'r':
+            info = opt.getptrinfo(box)
+            if info is not None and info.is_virtual():
+                result = info.visitor_dispatch_virtual_type(self)
+                self.info[box] = result
+                info.visitor_walk_recursive(box, self, opt)
+                result.fieldstate = [self.create_state_or_none(b, opt)
+                                     for b in self.fieldboxes[box]]
+            else:
+                result = self.visit_not_virtual(box)
+                self.info[box] = result
+        elif box.type == 'i' or box.type == 'f':
+            result = self.visit_not_virtual(box)
+            self.info[box] = result
+        else:
+            assert False
+        return result
 
     def get_virtual_state(self, jump_args):
-        self.optimizer.force_at_end_of_preamble()
-        already_forced = {}
         if self.optimizer.optearlyforce:
             opt = self.optimizer.optearlyforce
         else:
             opt = self.optimizer
-        values = [self.getvalue(box).force_at_end_of_preamble(already_forced,
-                                                              opt)
-                  for box in jump_args]
+        state = []
+        self.info = {}
+        for box in jump_args:
+            state.append(self.create_state(box, opt))
+        return VirtualState(state)
 
-        for value in values:
-            value.visitor_walk_recursive(self)
-        return VirtualState([self.state(box) for box in jump_args])
+    def visit_not_virtual(self, box):
+        return NotVirtualStateInfo(self.optimizer.cpu, box.type,
+                                   self.optimizer.getinfo(box))
 
-    def visit_not_virtual(self, value):
-        is_opaque = value in self.optimizer.opaque_pointers
-        return NotVirtualStateInfo(value, is_opaque)
-
-    def visit_virtual(self, known_class, fielddescrs):
+    def visit_virtual(self, descr, fielddescrs):
+        known_class = ConstInt(descr.get_vtable())
         return VirtualStateInfo(known_class, fielddescrs)
 
     def visit_vstruct(self, typedescr, fielddescrs):
@@ -574,147 +692,5 @@ class VirtualStateConstructor(VirtualVisitor):
         # to be non-None, so clearing is not important any more
         return VArrayStateInfo(arraydescr)
 
-    def visit_varraystruct(self, arraydescr, fielddescrs):
-        return VArrayStructStateInfo(arraydescr, fielddescrs)
-
-
-class BoxNotProducable(Exception):
-    pass
-
-
-class ShortBoxes(object):
-    def __init__(self, optimizer, surviving_boxes):
-        self.potential_ops = {}
-        self.alternatives = {}
-        self.synthetic = {}
-        self.rename = {}
-        self.optimizer = optimizer
-        self.assumed_classes = {}
-
-        if surviving_boxes is not None:
-            for box in surviving_boxes:
-                self.potential_ops[box] = None
-            optimizer.produce_potential_short_preamble_ops(self)
-
-            self.short_boxes = {}
-            self.short_boxes_in_production = {}
-
-            for box in self.potential_ops.keys():
-                try:
-                    self.produce_short_preamble_box(box)
-                except BoxNotProducable:
-                    pass
-
-            self.short_boxes_in_production = None # Not needed anymore
-        else:
-            self.short_boxes = {}
-
-    def prioritized_alternatives(self, box):
-        if box not in self.alternatives:
-            return [self.potential_ops[box]]
-        alts = self.alternatives[box]
-        hi, lo = 0, len(alts) - 1
-        while hi < lo:
-            if alts[lo] is None: # Inputarg, lowest priority
-                alts[lo], alts[-1] = alts[-1], alts[lo]
-                lo -= 1
-            elif alts[lo] not in self.synthetic: # Hi priority
-                alts[hi], alts[lo] = alts[lo], alts[hi]
-                hi += 1
-            else: # Low priority
-                lo -= 1
-        return alts
-
-    def renamed(self, box):
-        if box in self.rename:
-            return self.rename[box]
-        return box
-
-    def add_to_short(self, box, op):
-        if op:
-            op = op.clone()
-            for i in range(op.numargs()):
-                op.setarg(i, self.renamed(op.getarg(i)))
-        if box in self.short_boxes:
-            if op is None:
-                oldop = self.short_boxes[box].clone()
-                oldres = oldop.result
-                newbox = oldop.result = oldres.clonebox()
-                self.rename[box] = newbox
-                self.short_boxes[box] = None
-                self.short_boxes[newbox] = oldop
-            else:
-                newop = op.clone()
-                newbox = newop.result = op.result.clonebox()
-                self.short_boxes[newop.result] = newop
-            value = self.optimizer.getvalue(box)
-            self.optimizer.emit_operation(ResOperation(rop.SAME_AS, [box], newbox))
-            self.optimizer.make_equal_to(newbox, value)
-        else:
-            self.short_boxes[box] = op
-
-    def produce_short_preamble_box(self, box):
-        if box in self.short_boxes:
-            return
-        if isinstance(box, Const):
-            return
-        if box in self.short_boxes_in_production:
-            raise BoxNotProducable
-        self.short_boxes_in_production[box] = None
-
-        if box in self.potential_ops:
-            ops = self.prioritized_alternatives(box)
-            produced_one = False
-            for op in ops:
-                try:
-                    if op:
-                        for arg in op.getarglist():
-                            self.produce_short_preamble_box(arg)
-                except BoxNotProducable:
-                    pass
-                else:
-                    produced_one = True
-                    self.add_to_short(box, op)
-            if not produced_one:
-                raise BoxNotProducable
-        else:
-            raise BoxNotProducable
-
-    def add_potential(self, op, synthetic=False):
-        if op.result and op.result in self.optimizer.values:
-            value = self.optimizer.values[op.result]
-            if value in self.optimizer.opaque_pointers:
-                classbox = value.get_constant_class(self.optimizer.cpu)
-                if classbox:
-                    self.assumed_classes[op.result] = classbox
-        if op.result not in self.potential_ops:
-            self.potential_ops[op.result] = op
-        else:
-            if op.result not in self.alternatives:
-                self.alternatives[op.result] = [self.potential_ops[op.result]]
-            self.alternatives[op.result].append(op)
-        if synthetic:
-            self.synthetic[op] = True
-
-    def debug_print(self, logops):
-        if 0:
-            debug_start('jit-short-boxes')
-            for box, op in self.short_boxes.items():
-                if op:
-                    debug_print(logops.repr_of_arg(box) + ': ' + logops.repr_of_resop(op))
-                else:
-                    debug_print(logops.repr_of_arg(box) + ': None')
-            debug_stop('jit-short-boxes')
-
-    def operations(self):
-        if not we_are_translated(): # For tests
-            ops = self.short_boxes.values()
-            ops.sort(key=str, reverse=True)
-            return ops
-        return self.short_boxes.values()
-
-    def producer(self, box):
-        return self.short_boxes[box]
-
-    def has_producer(self, box):
-        return box in self.short_boxes
+    def visit_varraystruct(self, arraydescr, length, fielddescrs):
+        return VArrayStructStateInfo(arraydescr, fielddescrs, length)
