@@ -406,7 +406,7 @@ class UsingFrameworkTest(object):
                 try:
                     g()
                 except:
-                    os.write(1, "hallo")
+                    pass #os.write(1, "hallo")
         def f1(i):
             if i:
                 raise TypeError
